Tip: Ctrl-x t vloží názvy označených souborů na příkazový řádek.

Tip: Ctrl-x p vloží na příkazový řádek název složky, ve které se nacházíte.

Tip: Doplňování: použijte Meta-Tab (nebo Esc+Tab). Dalším stiskem se vypíše seznam možností.

Tip: V historii příkazů se pohybuje pomocí Meta-p a Meta-n.

Tip: Chcete zadat speciální znak (např. tabulátor)? Použijte Ctrl-q a pak daný znak.

Tip: Nechcete už vidět tipy? Vypněte je v nabídce Nastavení/Rozvržení.

Tip: Složky vyberete přidáním lomítka na konec vzoru pro hledání shody v názvu.

Tip: Pokud váš terminál nezná nebo nemá funkční (F1-12) klávesy, použijte Esc+číslo.

Tip: Domovská stránka GNU Midnight Commander je: http://www.midnight-commander.org/

Tip: Chyby hlaste na mc-devel@gnome.org (v angličtině).

Tip: Pomocí klávesy Tab přejdete na další panel.

Tip: Stiskem Enter na souboru .tar nebo .tar.gz si prohlédnete jeho obsah.

Tip: Existuje také manuálová stránka [mc(1)].

Tip: Přejete si klávesy stejně jako v prohlížeči lynx? Přepněte si to v Nastavení.

Tip: Makra uvozená % fungují i na příkazovém řádku.

Tip: Pomocí Meta-! můžete spustit příkaz a prohlédnout si výstup v prohlížeči.

Tip: Zobrazení v panelech může být upraveno. Z „man mc“ se dozvíte více.

Tip: %D/%T se rozvine na soubory označené v druhém panelu.

Tip: Chcete opět váš shell? Stiskněte Ctrl-o. Do MC se vrátíte opět přes Ctrl-o.

Tip: Nastavením proměnné CDPATH si můžete ušetřit psaní na klávesnici.

Tip: Pokud chcete vidět soubory začínající tečkou, vyjádřete to v Nastavení.

Tip: Chcete vidět zálohy svých souborů (končí ~)? Použijte Nastavení.

Tip: Doplňování funguje ve vstupních řádcích všech dialogů. Použijte Meta-Tab.

Tip: Na terminálech s pomalou odezvou zkuste parametr -s (mc -s).

Tip: S nalezenými soubory je možné dále pracovat. Stiskněte tlačítko Panelizace.

Tip: Pro opravdu složité vyhledávání zkuste příkaz příkaz Externí panelizace.

Tip: Pokud chcete změnit složku při práci na příkazovém řádku, zkuste Meta-c.

Pozn.: Příkazy shellu nefungují, pokud nepracujete na lokálním systému souborů.

Tip: Smazaný text obnovíte Ctrl-y.

Tip: Pokud některé klávesy nefungují, použijte menu Nastavení/Učení kláves.

Tip: K zobrazení výstupu příkazu v okně použijte Meta-!

Tip: F13 (nebo Shift-F3) vyvolá prohlížení bez zpracování obsahu souboru.

Tip: Editor pro klávesu F4 určíte pomocí proměnné prostředí EDITOR.

Tip: Externí prohlížeč určíte proměnnými VIEWER nebo PAGER

Tip: Všechny žádosti o potvrzení je možné vypnout v Nastavení/Potvrzování.

Tip: K seznamu často používaných složek se můžete dostat i přes Ctrl-\.

Tip: Na anonymní FTP se připojíte pomocí „cd ftp://stroj.domena.cz“.

Tip: MC má vestavěný FTP. Podívejte se do nabídka Levý/FTP spojení na počítač.

Tip: Meta-t rychle mění formát zobrazení panelů.

Tip: Při FTP spojení můžete zadat i jméno: „cd ftp://jmeno@stroj.domena.cz“.

Tip: RPM soubory si prohlédnete stiskem Enter.

Tip: Pro označení složek v dialogovém okně výběru souborů, přidejte lomítko.

Tip: Chcete-li použít myš pro zkopírování nebo vložení textu, stiskněte Shift.

Tip: V seznamu složek můžete mít i často navštěvované FTP servery. Zkuste Ctrl-\.
