<style type="text/css">

body {
  padding: 1em 2.5em;
  line-height: 140%;
}

code, pre {
  background-color: #FFB;
  padding: 2px 3px;
}

h1, h2 { text-indent: -0.5em; }
h1, h2, h3 {
  color: #005A9C;
  font-family: sans-serif;
}

:target {
  background: linear-gradient(45deg, #FFF 0%, #7CC 100%);
}

</style>
